@@ -13,51 +13,136 @@ import warnings
 from .types import BaseType, ConcreteType, ListType, SetType, MapType, OptionalType, ImplicitUnionType
 from .utils import type_repr, unpack_type_hint, find_generic_bases, _ORIGIN_CONVERSION
 
-has_pep585_generics = not (sys.version_info < (3, 9))
-#has_pep604_union_types = not (sys.version_info < (3, 10))
+# has_pep585_generics = not (sys.version_info < (3, 9))
+# has_pep604_union_types = not (sys.version_info < (3, 10))
 
 
 @dataclasses.dataclass
 class TypeHintAdapterError(Exception):
   adapter: 'TypeHintAdapter'
   message: str
+  causes: t.List['TypeHintAdapterError'] = dataclasses.field(default_factory=list)
 
   def __str__(self) -> str:
-    return self.message
+    parts = [self.message] + [str(c) for c in self.causes]
+    return '\n'.join(parts)
 
 
-class ForwardReferenceResolver(abc.ABC):
+@dataclasses.dataclass
+class TypeContext:
+  """ Tracking context for type adaptation. """
 
-  @abc.abstractmethod
-  def get_type(self, forward_ref: str) -> t.Union[t.Any, t.Type]:
-    ...
+  root_adapter: 'TypeHintAdapter'
+  globals_: t.Optional[t.Mapping[str, t.Any]] = None
+  locals: t.Optional[t.Mapping[str, t.Any]] = None
+  type_vars: t.Optional[t.Mapping[t.TypeVar, t.Any]] = None
+  _source: t.Any = None
 
+  def adapt_type_hint(self, type_hint: t.Any) -> BaseType:
+    return self.root_adapter.adapt_type_hint(type_hint, self.with_type_vars_of(type_hint))
 
-class ModuleForwardReferenceResolver(ForwardReferenceResolver):
+  def resolve_forward_reference(self, forward_ref: t.Union[str, t.ForwardRef]) -> t.Any:
+    if isinstance(forward_ref, str):
+      forward_ref = t.ForwardRef(forward_ref)
+    # FIXME (@NiklasRosenstein): Relying on internal typing API here.
+    if sys.version_info < (3, 9):
+      return forward_ref._evaluate(self.globals_, self.locals)  # type: ignore
+    else:
+      return forward_ref._evaluate(self.globals_, self.locals, frozenset())  # type: ignore
 
-  def __init__(self, module: types.ModuleType) -> None:
-    self._module = module
+  def resolve_type_var(self, type_variable: t.TypeVar) -> t.Any:
+    try:
+      return (self.type_vars or {})[type_variable]
+    except KeyError:
+      raise RuntimeError(f'cannot resolve {type_variable!r} in {self._source!r}')
 
-  def get_type(self, forward_ref: str) -> t.Union[t.Any, t.Type]:
-    parts = forward_ref.split('.')
-    current = self._module
-    for part in parts:
-      try:
-        current = getattr(current, part)
-      except AttributeError:
-        raise KeyError(forward_ref)
-    return current
+  def apply_type_vars(self, type_hint: t.Any) -> t.Any:
+    if self.type_vars is None:
+      return type_hint
+    if not hasattr(type_hint, '__origin__') and hasattr(type_hint, '__parameters__'):
+      args = tuple(self.type_vars.get(k) for k in type_hint.__parameters__)
+      if args:
+        return type_hint[args]
+    return type_hint
+
+  def with_scope_of(self, type_hint: t.Any) -> 'TypeContext':
+    """ Tries to assume the scope (i.e. globals and locals) of the type given with *type_hint* and returns a new
+    type context. If no scope can be identified for the type hint, the returned context will not contain a scope.
+    """
+
+    if hasattr(type_hint, '__origin__'):
+      type_hint = type_hint.__origin__
+
+    if isinstance(type_hint, type):
+      module = sys.modules.get(type_hint.__module__)
+    else:
+      module = None
+
+    return TypeContext(
+      root_adapter=self.root_adapter,
+      globals_=vars(module) if module else self.globals_,
+      locals=None,
+      type_vars=self.type_vars,
+      _source=type_hint,
+    )
+
+  def with_type_vars_of(self, type_hint: t.Any, l: bool = False) -> 'TypeContext':
+    """ Extracts assigned type paremtrizations and carries them forward in a new type context returned by this
+    function. """
+
+    def _resolve_type_var(tv: t.TypeVar) -> t.Any:
+      if self.type_vars and tv in self.type_vars:
+        return self.type_vars[tv]
+      elif isinstance(tv.__bound__, (str, t.ForwardRef)):
+        return self.resolve_forward_reference(tv.__bound__)
+      elif isinstance(tv.__bound__, str):
+        return self.resolve_forward_reference(tv.__bound__)
+      elif tv.__bound__:
+        return tv.__bound__
+      else:
+        return t.Any
+
+    def _evaluate(a: t.Any) -> t.Any:
+      if isinstance(a, t.TypeVar):
+        return _resolve_type_var(a)
+      elif isinstance(a, t.ForwardRef):
+        return self.resolve_forward_reference(a)
+      return a
+
+    if hasattr(type_hint, '__origin__') and hasattr(type_hint.__origin__, '__parameters__'):
+      args = [_evaluate(a) for a in type_hint.__args__]
+      type_hint = type_hint.__origin__
+      type_vars = dict(zip(type_hint.__parameters__, args))
+
+    elif hasattr(type_hint, '__parameters__'):
+      type_vars = {p: _resolve_type_var(p) for p in type_hint.__parameters__}
+
+    else:
+      type_vars = {}
+
+    if hasattr(type_hint, '__orig_bases__'):
+      scoped = self.with_scope_of(type_hint)
+      for base in type_hint.__orig_bases__:
+        type_vars = {**(scoped.with_type_vars_of(base).type_vars or {}), **type_vars}
+
+    if self.type_vars:
+      type_vars = {**self.type_vars, **type_vars}
+
+    return TypeContext(
+      root_adapter=self.root_adapter,
+      globals_=self.globals_,
+      locals=self.locals,
+      type_vars=type_vars,
+      _source=type_hint,
+    )
 
 
 class TypeHintAdapter(abc.ABC):
 
   Error = TypeHintAdapterError
 
-  def adapt_type_hint(self, type_hint: t.Any, recurse: t.Optional['TypeHintAdapter'] = None, resolver: t.Optional[ForwardReferenceResolver] = None) -> BaseType:
-    return self._adapt_type_hint_impl(type_hint, recurse or self, resolver)
-
   @abc.abstractmethod
-  def _adapt_type_hint_impl(self, type_hint: t.Any, recurse: 'TypeHintAdapter', resolver: t.Optional[ForwardReferenceResolver]) -> BaseType: ...
+  def adapt_type_hint(self, type_hint: t.Any, context: TypeContext) -> BaseType: ...
 
 
 class DefaultTypeHintAdapter(TypeHintAdapter):
@@ -65,15 +150,9 @@ class DefaultTypeHintAdapter(TypeHintAdapter):
   Adapter for all the supported standard #typing type hints.
   """
 
-  def _adapt_type_hint_impl(self, type_hint: t.Any, recurse: TypeHintAdapter, resolver: t.Optional[ForwardReferenceResolver]) -> BaseType:
+  def adapt_type_hint(self, type_hint: t.Any, context: TypeContext) -> BaseType:
     generic, args = unpack_type_hint(type_hint)
-    from_typing = lambda th: recurse._adapt_type_hint_impl(th, recurse, resolver)
-
-    if has_pep585_generics and isinstance(type_hint, types.GenericAlias):  # type: ignore
-      if not resolver and any(isinstance(x, str) for x in type_hint.__args__):
-        raise ValueError(
-          f'encountered forward reference in PEP585 generic `{type_hint}` but no ForwardReferenceResolver is supplied'
-        )
+    from_typing = context.adapt_type_hint
 
     # Support custom subclasses of typing generic aliases (e.g. class MyList(t.List[int])
     # or class MyDict(t.Mapping[K, V])). If we find a type like that, we keep a reference
@@ -120,12 +199,7 @@ class DefaultTypeHintAdapter(TypeHintAdapter):
         return from_typing(type_)
 
     if isinstance(type_hint, str):
-      if resolver:
-        try:
-          return from_typing(resolver.get_type(type_hint))
-        except KeyError:
-          pass
-      raise TypeHintAdapterError(self, 'cannot resolve forward reference {type_hint!r}')
+      return from_typing(context.resolve_forward_reference(type_hint))
 
     if isinstance(type_hint, type):
       return from_typing(ConcreteType(type_hint))
@@ -133,6 +207,13 @@ class DefaultTypeHintAdapter(TypeHintAdapter):
     if isinstance(type_hint, BaseType):
       return type_hint
 
+    if isinstance(type_hint, t.TypeVar):
+      resolved = context.resolve_type_var(type_hint)
+      return context.with_scope_of(resolved).adapt_type_hint(resolved)
+
+    if generic is not None and isinstance(generic, type):  # Probably a subclass of typing.Generic
+      return context.root_adapter.adapt_type_hint(generic, context)
+
     raise TypeHintAdapterError(self, f'unsupported type hint {type_hint!r}')
 
 
@@ -172,12 +253,12 @@ class ChainTypeHintAdapter(TypeHintAdapter):
       item = (priority, self._priorities[priority])
       self._ordered.insert(bisect.bisect(self._ordered, item), item)
 
-  def _adapt_type_hint_impl(self, type_hint: t.Any, recurse: TypeHintAdapter, resolver: t.Optional[ForwardReferenceResolver]) -> BaseType:
+  def adapt_type_hint(self, type_hint: t.Any, context: TypeContext) -> BaseType:
     errors = []
     for _priority_group, adapters in self._ordered:
       for adapter in adapters:
         try:
-          type_hint = adapter.adapt_type_hint(type_hint, recurse, resolver)
+          type_hint = adapter.adapt_type_hint(type_hint, context)
         except TypeHintAdapterError as exc:
           errors.append(exc)
         if isinstance(type_hint, BaseType) and any(x(type_hint) for x in self._stop_conditions):
@@ -188,4 +269,4 @@ class ChainTypeHintAdapter(TypeHintAdapter):
       raise TypeHintAdapterError(self, 'no adapters registered')
     summary = '\n'.join(f'{str(exc.adapter)}: {exc.message}' for exc in errors)
     summary = textwrap.indent(summary, '  ')
-    raise TypeHintAdapterError(self, 'no available adapter matched\n' + summary)
+    raise TypeHintAdapterError(self, 'no available adapter matched\n' + summary, errors)
